/* Arduino "guitar" based on Mozzi audio library by Tim Barrass.
 *
 * GUITARRON VERSION!
 *
 * Code by Andrew McPherson, Queen Mary University of London, September 2012
 *
 * http://www.eecs.qmul.ac.uk/
 *
 * Circuit: 
 *   audio output on pin 9 (to filter and LM386 amplifier)
 *   analog inputs (knobs/sliders) on pins A0-A3
 *   digital inputs (buttons) on pins 2-5
 */ 
 
#include <MozziGuts.h>
#include <Oscil.h>
#include <tables/sin2048_int8.h>
#include <tables/waveshape_chebyshev_3rd_256_int8.h>
#include <tables/waveshape_chebyshev_6th_256_int8.h>
#include <tables/waveshape_compress_512_to_488_int16.h>
#include <tables/waveshape2_softerclip_int8.h>
#include <tables/saw_analogue512_int8.h>
#include <tables/smoothsquare8192_int8.h>
#include <MozziGuts.h>
#include <Oscil.h>
#include <WaveShaper.h>
#include <EventDelay.h>
#include <utils.h>
#include <Smooth.h>
#include <fixedMath.h> // for fractional modulation frequency
#include <Ead.h> // exponential attack decay

#define CONTROL_RATE 128 // powers of 2 please

// Pin definitions for analog and digital inputs:

#define PITCH_PIN A0
#define VOLUME_PIN A1
#define DECAY_PIN A2
#define TUNING_PIN A3

#define STRING1_PIN 2
#define STRING2_PIN 3
#define STRING3_PIN 4
#define STRING4_PIN 5

// Basic pitch of the instrument (MIDI 45 = A2)
#define MIDI_NOTE_MIN (Q16n16)(28L << 16L)

Oscil <SMOOTHSQUARE8192_NUM_CELLS, AUDIO_RATE> aString1(SMOOTHSQUARE8192_DATA); // sine wave sound source
Oscil <SMOOTHSQUARE8192_NUM_CELLS, AUDIO_RATE> aString2(SMOOTHSQUARE8192_DATA);
Oscil <SMOOTHSQUARE8192_NUM_CELLS, AUDIO_RATE> aString3(SMOOTHSQUARE8192_DATA);
Oscil <SMOOTHSQUARE8192_NUM_CELLS, AUDIO_RATE> aString4(SMOOTHSQUARE8192_DATA);

WaveShaper <int> aCompress(WAVESHAPE_COMPRESS_512_TO_488_DATA); // to compress instead of dividing by 2 after adding signals

Ead kEnvelopeString1(CONTROL_RATE);
Ead kEnvelopeString2(CONTROL_RATE);
Ead kEnvelopeString3(CONTROL_RATE);
Ead kEnvelopeString4(CONTROL_RATE);

/* Parameters controlled by analog inputs */
unsigned long pitch;
unsigned char volume;
unsigned int decay;
unsigned int tuning;

unsigned char string1Val = 0, string2Val = 0, string3Val = 0, string4Val = 0;
unsigned char lastString1Val = 0, lastString2Val = 0, lastString3Val = 0, lastString4Val = 0;
unsigned int attackMilliseconds = 20, decayMilliseconds = 3000;
unsigned int lastDecayValue = 0;
int gain1, gain2, gain3, gain4;

const unsigned char blackKeys[12] = {0, 1, 0, 1, 0, 0, 1, 0, 1, 0, 1, 0};

void updateAttackDecayValues() {
  kEnvelopeString1.set(attackMilliseconds, decayMilliseconds);
  kEnvelopeString2.set(attackMilliseconds, decayMilliseconds);
  kEnvelopeString3.set(attackMilliseconds, decayMilliseconds);
  kEnvelopeString4.set(attackMilliseconds, decayMilliseconds);  
}

void setup(){
  startMozzi(CONTROL_RATE);
  
  aString1.setFreq(100u); // set the frequency with an unsigned int or a float
  aString2.setFreq(100u);
  aString3.setFreq(100u);
  aString4.setFreq(100u);
  
  updateAttackDecayValues();
  setupFastAnalogRead();
  
  Serial.begin(57600); 
}

void updateControl() {
  static Q16n16 midiNote, midiNoteInt, baseFrequency;
  static long midiNoteFrac;
  
  /* Read sensor data and update audio parameters accordingly */
  pitch = analogRead(PITCH_PIN);  // Range: 0 to 1023
  volume = analogRead(VOLUME_PIN) >> 2;  // Range: 0 to 255
  decay = analogRead(DECAY_PIN);  // Range: 0 to 1023
  tuning = analogRead(TUNING_PIN); // Range: 0 to 1023
  
  /* Read digital sensor data (buttons) */
  string1Val = digitalRead(STRING1_PIN);
  string2Val = digitalRead(STRING2_PIN);
  string3Val = digitalRead(STRING3_PIN);
  string4Val = digitalRead(STRING4_PIN);
  
  /* Only change the decay value if a significant change has occurred,
   * to filter out noise on the input. This will reduce unneeded calculations. */
  if(abs(decay - lastDecayValue) > 4) {
    decayMilliseconds = decay << 2; 
    lastDecayValue = decay; 
    updateAttackDecayValues();
  }
  
  /* Values that are high now and were low last cycle mean that
   * a button has been pushed. Start the corresponding string. */
  if(string1Val == HIGH && lastString1Val == LOW) {
    kEnvelopeString1.start();
  }
  if(string2Val == HIGH && lastString2Val == LOW) {
    kEnvelopeString2.start();
  }
  if(string3Val == HIGH && lastString3Val == LOW) {
    kEnvelopeString3.start();
  }
  if(string4Val == HIGH && lastString4Val == LOW) {
    kEnvelopeString4.start();
  }
  
  lastString1Val = string1Val;
  lastString2Val = string2Val;
  lastString3Val = string3Val;
  lastString4Val = string4Val;
  
  /* pitch: 0 = minimum value; 1024 = +32 semitones (~2.5 octaves) 
   * 6 bits left to get to decimal point + 5 bits more = 11 bit shift */
  
  midiNote = MIDI_NOTE_MIN + (Q16n16)(pitch << 9);
  
  /* Set the amount the pitch is quantized. */
  midiNoteInt = (midiNote + 0x00008000) & 0xFFFF0000; // Nearest MIDI note (Q16n16 rounding)
  midiNoteFrac = (long)midiNote - (long)midiNoteInt; /* Fractional part only */
  midiNoteFrac = (midiNoteFrac * (1023 - tuning)) >> 10; /* Scale the fractional part */
  midiNote = midiNoteInt + midiNoteFrac;
  
  baseFrequency = Q16n16_mtof(midiNote);
  
  /* Calculate the ratios for each of the strings to create a harmony.
   * Ratios:
   * Octave      --> 2
   * Fifth       --> 3/2
   * Fourth      --> 4/3
   * Major Third --> 5/4
   * Minor Third --> 6/5
   * Major 2nd   --> 9/8
   * Tritone     --> sqrt(2) or ~45/32
   */
   
  aString1.setFreq_Q16n16(baseFrequency);
  aString2.setFreq_Q16n16(baseFrequency * 4 / 3);
  aString3.setFreq_Q16n16(baseFrequency * 16 / 9);
  aString4.setFreq_Q16n16(baseFrequency * 64 / 27);
  
  gain1 = kEnvelopeString1.next();
  gain2 = kEnvelopeString2.next();
  gain3 = kEnvelopeString3.next();
  gain4 = kEnvelopeString4.next();
  
  if(gain1 > 96) gain1 = 96;
  if(gain2 > 96) gain2 = 96;
  if(gain3 > 96) gain3 = 96;
  if(gain4 > 96) gain4 = 96;
}

// Given a MIDI input pitch, round it to the nearest note of a
// major scale starting at MIDI_NOTE_MIN.
Q16n16 restrictMidiToMajorScale(Q16n16 input) {
  int pitchClass = ((input - MIDI_NOTE_MIN) >> 16) % 12;
  
  // If it's a chromatic pitch (black key), return the note below
  // (the nearest white key), otherwise return the current note.
  if(blackKeys[pitchClass]) {
    return input - (1L << 16);  
  }
  else {
    return input;  
  }
}


int updateAudio(){
  char aOut1 = 0, aOut2 = 0, aOut3 = 0, aOut4 = 0;
 
  aOut1 = aString1.next();
  aOut2 = aString2.next();
  aOut3 = aString3.next(); 
  aOut4 = aString4.next();

  long waveSum = (gain1*aOut1 + gain2*aOut2 + gain3*aOut3 + gain4*aOut4) >> 8;
  waveSum = (waveSum * volume) >> 8;

  // use a waveshaping table to squeeze 2 summed 8 bit signals into the range -244 to 243
  int waveshapedOut = aCompress.next(256u + waveSum);
  
  return waveshapedOut;
}


void loop(){
  audioHook();
}




